# -*- coding: utf-8 -*-

"""
Package to parse logical tag expressions
"""

import pytest

from tagexpressions import parse


@pytest.mark.parametrize('infix, expected', [
    ('a and b', '( a and b )'),
    ('a or b', '( a or b )'),
    ('not a', 'not ( a )'),
    ('( a and b ) or ( c and d )', '( ( a and b ) or ( c and d ) )'),
    ('not a or b and not c or not d or e and f', '( ( ( not ( a ) or ( b and not ( c ) ) ) or not ( d ) ) or ( e and f ) )')
])
def test_parser(infix, expected):
    """Test the tag expression parser"""
    assert str(parse(infix)) == expected


@pytest.mark.parametrize('infix, values, expected', [
    ('a and b', ['a', 'b'], True),
    ('a and b', ['b', 'a'], True),
    ('a and b', ['a'], False),
    ('a or b', ['a', 'b'], True),
    ('a or b', ['b', 'a'], True),
    ('a or b', ['a'], True),
    ('a or b', ['b'], True),
    ('a or b', ['c'], False),
    ('not a', ['b'], True),
    ('not a', ['a'], False),
    ('not a', [], True)
])
def test_basic_evaluation(infix, values, expected):
    """Test basic tag expression evaluation"""
    assert parse(infix).evaluate(values) == expected


@pytest.mark.parametrize('infix, values, expected', [
    ('( a and b ) or ( c and d )', ['a', 'b', 'c', 'd'], True),
    ('( a and b ) or ( c and d )', ['c', 'd'], True),
    ('( a and b ) or ( c and d )', ['a', 'b'], True),
    ('( a and b ) or ( c and d )', ['a', 'c'], False),
    ('not a or b and not c or not d or e and f', ['b', 'e', 'f'], True),
    ('not a or b and not c or not d or e and f', ['a', 'b', 'e', 'f'], True),
    ('not a or b and not c or not d or e and f', ['a', 'b', 'c'], True),
    ('not a or b and not c or not d or e and f', ['a', 'c', 'd'], False),
])
def test_complex_evaluation(infix, values, expected):
    """Test complex tag expression evaluation"""
    assert parse(infix).evaluate(values) == expected



# -*- coding: utf-8 -*-

"""
Package to parse logical tag expressions
"""

class Literal(object):  # pylint: disable=too-few-public-methods
    """
    Represents a literal expression
    """
    def __init__(self, value):
        self.value = value

    def evaluate(self, values):
        """Evaluate the expression

        Check if the literal is in the given
        set of values.
        """
        return self.value in values

    def __str__(self):
        return self.value


class Or(object):  # pylint: disable=too-few-public-methods
    """
    Represents an "OR" expression
    """
    def __init__(self, left, right):
        self.left = left
        self.right = right

    def evaluate(self, values):
        """Evaluate the "OR" expression

        Check if the left "or" right expression
        evaluate to True.
        """
        return self.left.evaluate(values) or self.right.evaluate(values)

    def __str__(self):
        return '( {0} or {1} )'.format(self.left, self.right)


class And(object):  # pylint: disable=too-few-public-methods
    """
    Represent an "AND" expression
    """
    def __init__(self, left, right):
        self.left = left
        self.right = right

    def evaluate(self, values):
        """Evaluate the "AND" expression

        Check if the left "and" right expression
        evaluate to True.
        """
        return self.left.evaluate(values) and self.right.evaluate(values)

    def __str__(self):
        return '( {0} and {1} )'.format(self.left, self.right)


class Not(object):  # pylint: disable=too-few-public-methods
    """
    Represents a "NOT" expression
    """
    def __init__(self, expression):
        self.expression = expression

    def evaluate(self, values):
        """
        Evaluate the "NOT" expression

        Check if the given expression does not
        evaluate to True.
        """
        return not self.expression.evaluate(values)

    def __str__(self):
        return 'not ( {0} )'.format(self.expression)



# -*- coding: utf-8 -*-

import ast
import os
import codecs
from setuptools import setup, find_packages

"""
Package to parse logical tag expressions
"""

PROJECT_ROOT = os.path.dirname(__file__)


class VersionFinder(ast.NodeVisitor):
    def __init__(self):
        self.version = None

    def visit_Assign(self, node):
        try:
            if node.targets[0].id == 'version':
                self.version = node.value.s
        except:
            pass


def read_version():
    """Read version from tagexpressions/__init__.py without loading any files"""
    finder = VersionFinder()
    path = os.path.join(PROJECT_ROOT, 'tagexpressions', '__init__.py')
    with codecs.open(path, 'r', encoding='utf-8') as fp:
        file_data = fp.read().encode('utf-8')
        finder.visit(ast.parse(file_data))

    return finder.version


def local_text_file(*f):
    path = os.path.join(PROJECT_ROOT, *f)
    with open(path, 'rt') as fp:
        file_data = fp.read()

    return file_data


def read_readme():
    """Read README content.
    If the README.rst file does not exist yet
    (this is the case when not releasing)
    only the short description is returned.
    """
    try:
        return local_text_file('README.rst')
    except IOError:
        return __doc__


setup(name='tag-expressions',
      version=read_version(),
      description=__doc__,
      long_description=read_readme(),
      url='http://github.com/timofurrer/tag-expressions',
      author='Timo Furrer',
      author_email='tuxtimo@gmail.com',
      include_package_data=True,
      packages=find_packages(exclude=['*tests*']),
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Environment :: Console',
          'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
          'Operating System :: MacOS :: MacOS X',
          'Operating System :: POSIX',
          'Operating System :: POSIX :: Linux',
          'Programming Language :: Python',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.3',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: Implementation',
          'Programming Language :: Python :: Implementation :: CPython',
          'Programming Language :: Python :: Implementation :: PyPy',
          'Topic :: Software Development :: Testing'
      ]
)



# -*- coding: utf-8 -*-

"""
Package to parse logical tag expressions
"""

from collections import deque

from .models import Literal, Or, And, Not

#: Holds the possible token associativities
ASSOC_LEFT = 0
ASSOC_RIGHT = 1

TOKEN_ASSOCS = {'or': ASSOC_LEFT, 'and': ASSOC_LEFT, 'not': ASSOC_RIGHT}

#: Holds the token precedences
TOKEN_PRECS = {'(': -2, ')': -1, 'or': 0, 'and': 1, 'not': 2}


class TagExpressionError(Exception):
    """Raised for tag expression specific errors"""
    pass


def parse(infix):
    """
    Parse the given string logical tag expression.
    """
    tokens = infix.replace('(', ' ( ').replace(')', ' ) ').strip().split()

    #: Holds the parsed operations
    ops = deque()
    #: Holds the parsed expressions
    expressions = deque()

    for token in tokens:
        if token in TOKEN_ASSOCS:
            while len(ops) > 0 and ops[-1] in TOKEN_ASSOCS and (
                (TOKEN_ASSOCS[token] == ASSOC_LEFT and TOKEN_PRECS[token] <= TOKEN_PRECS[ops[-1]]) or \
                  (TOKEN_ASSOCS[token] == ASSOC_RIGHT and TOKEN_PRECS[token] < TOKEN_PRECS[ops[-1]])):
                create_and_push_expression(ops.pop(), expressions)

            ops.append(token)
        elif token == '(':
            ops.append(token)
        elif token == ')':
            while len(ops) > 0 and ops[-1] != '(':
                create_and_push_expression(ops.pop(), expressions)

            if len(ops) == 0:
                raise TagExpressionError('Unclosed (')

            if ops[-1] == '(':
                ops.pop()
        else:
            create_and_push_expression(token, expressions)

    while len(ops) > 0:
        if ops[-1] == '(':
            raise TagExpressionError('Unclosed )')
        create_and_push_expression(ops.pop(), expressions)

    expression = expressions.pop()
    if len(expressions) > 0:
        raise TagExpressionError('Not empty')

    return expression


def create_and_push_expression(token, expressions):
    """Creates an expression from the given token and adds it
    to the stack of the given expression.

    In the case of "and" and "or" expressions the last expression
    is poped from the expression stack to link it to the new
    created one.
    """
    if token == 'and':
        right_expr = expressions.pop()
        expressions.append(And(expressions.pop(), right_expr))
    elif token == 'or':
        right_expr = expressions.pop()
        expressions.append(Or(expressions.pop(), right_expr))
    elif token == 'not':
        expressions.append(Not(expressions.pop()))
    else:
        expressions.append(Literal(token))
